/***********************************************************************
calcseq.c

	Calculation utility for siseq programs 
  
 	Naoki Sato   Nov. 28, 1999. Started initial project.
	
	Modified c_freq_i. Dec. 1, 1999.
	Added EraseCodingRegion. Dec. 4, 1999.
	Added SelectNonCoding. Dec. 4, 1999.
	Added PrintScTable and CountSimilarity. Dec. 16, 1999.
	Added MatchSeq. Feb. 18, 2000. 
	Added c_comp. May 8, 2000. Modified July 7, 2000.
	Added getProtMW. July 24, 2000.
	Added CountSkew, CountGC. Modified c_basecomp 
		to use c_basecomp_seq. Nov. 26, 2000
	Added codon_usage_comp. Jan. 15, 2003.
	Added MatchSeqN. Feb. 9, 2004.
	Added RemoveGaps and SelectSitesByInf. March 7, 2004.
	Added DumpSites. Oct. 1, 2017.
	
		--- Do not delete this comment. ---
 ***********************************************************************/

#include "Dmc.h"
#include "getseq2.h"
#include <math.h>
#include "molwt.h"

/* Prototype declaration */

int DumpSites(char *fname2, Seqfile **multiseq);
int RemoveGaps(Seqfile **multiseq,double level,int option);
int SelectSitesByInf(Seqfile **multiseq,double min, double max, enum data_type dtype1);
double InfContent(char *seq, long seqlen, enum data_type dtype1);
double getProtMW(Seqfile *protseq);
int MatchSeqN(char **seq1,char **seq2,char *seq_out,long seqlen_out,long match_len,double c,char *codon_table);
long MatchSeq(char *seq1,char *seq2,double c);
int alnStat(char *fname1, char *fname2, Seqfile **seqM, double thr);
int alnCorrect(int mode,char *fname1,char *fname2,Seqfile **seqM,double thr,char *nuc_file,Seqfile **seq_nuc,char *org_name,long offset);
int massAlnCorrect(char *fname1,char *fname2,double thr,char *nuc_file,long offset);
int UnequalLength(Seqfile **seqM);
void PrintScTable(FILE *fout,long N0,SimilarityCounts **sc1,int sw);
int CountSimilarity(Seqfile *seq1,Seqfile *seq2,long N0,double st[30][30],SimilarityCounts **sc1,long pos);
int EraseCodingRegion(Seqfile *seq_in,Seqfile **seq_nuc,int c,Boolean retain_intron);
int CountNonCoding(Seqfile *seq_in,long *NumSeq,int c);
int SelectNonCoding(Seqfile *seq_in,Seqfile **seq_out,long *NumSeq,int c);
int c_basecomp(Seqfile *seq_in,double basecomp[4]);
int c_freq(Seqfile *seq_in,long *freq,long maxdist,int modeM);
int c_freq_i(Seqfile *seq_in,long *freq,long i,int modeM);
int c_comp(Seqfile *seq_in, long code_count[30]);
int c_comp_char(char *seq,long seqlen,long code_count[30]);
int codon_usage_comp(Seqfile *seq_in, long *code_count);
void initialize_comp(long code_count[30]);
void initialize_comp2(long *code_count,unsigned dim);
int c_basecomp_seq(char *seq,double basecomp[4]);
double CountGCSkew(char *seq);
double CountATSkew(char *seq);
double CountGC(char *seq);

/* External functions */
#ifndef __CALCSEQ
#define __CALCSEQ
#endif

#include "libSeq.h"

/* Global variables  */

extern char aaname[20][2][10];
extern Boolean circular;
extern char *data_type_name[];
extern long num_seq;


/*********************************
Definition: 0=T, 1=C, 2=A, 3=G.
*********************************/

/***********************************************************/
int DumpSites(char *fname2, Seqfile **multiseq)
/***********************************************************/
{
	FILE *fout;
	long i, j;
	char *column;
	long seqlen = multiseq[0]->seqlen;

    if((column=(char*)calloc(num_seq + 2,sizeof(char)))==NULL){
        fprintf(stderr,"Memory allocation error in DumpSites 1.\n");
        return TRUE;
    }

	if((fout=fopen(fname2,"w"))==NULL){
		fprintf(stderr,"Cannot open file: %s\n\n",fname2);
		return TRUE;
	}

	fprintf(fout,"%ld sites, %ld sequences\n",seqlen,num_seq);
	for(j=0;j<seqlen;j++){
		for(i=0;i<num_seq;i++){
			column[i] = multiseq[i]->seq[j];	
		}
		column[num_seq] = '\0';
		fprintf(fout, "%s\n",column);
	}

	fclose(fout);
	return FALSE;
}
/***********************************************************/

/***********************************************************/
double CountGC(char *seq)
/* This function returns GC content.
************************************************************/
{
	double GCcont=0.0;
	double basecomp[4];

	if(c_basecomp_seq(seq,basecomp)){
		fprintf(stderr,"Error in c_basecomp_seq in CountSkew.\n");
		return 999;
	}

	GCcont=(basecomp[3] + basecomp[1]);
	return GCcont;
}

/***********************************************************/
double CountGCSkew(char *seq)
/* This function returns GC skew which is defined by
   (G - C)/(G + C)
************************************************************/
{
	double skew=0.0;
	double basecomp[4];

	if(c_basecomp_seq(seq,basecomp)){
		fprintf(stderr,"Error in c_basecomp_seq in CountGCSkew.\n");
		return 999;
	}

	if(basecomp[3]+basecomp[1]<=0) {
		fprintf(stderr,"Strange composition G+C in CountGCSkew.\n");
		return 999;
	}

	if((basecomp[3] + basecomp[1]) <= 0.01) return 0.0;
	skew=(basecomp[3] - basecomp[1])/(basecomp[3] + basecomp[1]);

	return skew;
}

/***********************************************************/
double CountATSkew(char *seq)
/* This function returns AT skew which is defined by
   (A - T)/(A + T)
************************************************************/
{
	double skew=0.0;
	double basecomp[4];

	if(c_basecomp_seq(seq,basecomp)){
		fprintf(stderr,"Error in c_basecomp_seq in CountATSkew.\n");
		return 999;
	}

	if(basecomp[2]+basecomp[0]<=0) {
		fprintf(stderr,"Strange composition A+T in CountATSkew.\n");
		return 999;
	}

	if((basecomp[2] + basecomp[0]) <= 0.01) return 0.0;
	skew=(basecomp[2] - basecomp[0])/(basecomp[2] + basecomp[0]);

	return skew;
}

/***********************************************************/
double getProtMW(Seqfile *protseq)
/* This function returns protein molecular weight. 
  On error, it returns 0.0. */
/***********************************************************/
{
	double mw=0.0;
	char aares;
	int i;
	char *p;	/* pointer to aares */

	if(protseq==NULL) return 0.0;
	if(protseq->seqlen < 2L) return 0.0;
	if(strlen(protseq->seq)<2) return 0.0;

	for(p=protseq->seq;*p!='\0';p++){
		if(islower(*p)) aares=toupper(*p);
		else if(isupper(*p)) aares=*p;
		else return 0.0;

		for(i=0;i<21;i++){
			if(aaname[i][1][0]==aares) break;
		}
		if(i>=0 && i<21){
			mw+=aawt[i];
		}
		else return 0.0;
	}

	mw=mw-H2O*(protseq->seqlen-1);	
	return mw;

}
/***********************************************************/

/***********************************************************/
int RemoveGaps(Seqfile **multiseq, double level, int option)
/* In a multiple alignment, sites that contain gaps at a level
higher than the given level are removed.
Returns TRUE if error occurs. */
/* Jan. 2008: Option 0 (default), remove gaps.
                     1, fill the gaps with bars
                     2, prints conserved sites 
                     3, fill the gaps with bars & outputs Eco table */
/***********************************************************/
{
	long i1,i2,j,i3,i4,j_eco;
	long gap_count;
	double gap_content;
	long seqlen = multiseq[0]->seqlen;

	if(option != 1 && option != 2 && option != 3) option = 0;

	if(level < 0.0){
		fprintf(stderr,"Negative level setting in RemoveGaps.\n");
		return TRUE;
	}
	if(option == 2 || option == 3) printf("\n");

	if(option == 3){
		for(j=0;j<num_seq;j++){
			if(strncmp(multiseq[j]->ID,"Eco",3)==0) break;
		}
		if(j==num_seq){
			fprintf(stderr,"Sequence Eco should be present.\n");
			return TRUE;
		} else {
			j_eco = j;
		}
	}
		 
	i3 = i4 = 1;		/* counts for Eco sequence and valid sequence */
	for(i1=0,i2=0;i2<seqlen;i2++){
		if(option == 3) printf("\n%ld\t",i2+1);

		gap_count = 0;
		for(j=0;j<num_seq;j++){
			if(multiseq[j]->seq[i2] == '-') gap_count += 1;
		}
		gap_content = (double)gap_count / (double)num_seq;
		if(gap_content <= level){
			for(j=0;j<num_seq;j++){
				if(option == 1 || option == 3) {
					if(islower(multiseq[j]->seq[i2])) multiseq[j]->seq[i1] = toupper(multiseq[j]->seq[i2]);
				} else multiseq[j]->seq[i1] = multiseq[j]->seq[i2];
			}
			if(option == 3) {
				if(multiseq[j_eco]->seq[i1] == '-') printf("-\t-\t%ld",i4++);
				else printf("%c\t%ld\t%ld",multiseq[j_eco]->seq[i1],i3++,i4++);
			}
			i1++;
			if(option == 2) printf("\n%ld",i2+1);
		} else {
			if(option == 1 || option == 3) {
				for(j=0;j<num_seq;j++){
					if(multiseq[j]->seq[i2] == '-') multiseq[j]->seq[i1] = '|';
					else if(isupper(multiseq[j]->seq[i2])) {
						multiseq[j]->seq[i1] = tolower(multiseq[j]->seq[i2]);
					}
				}
				
				if(option == 3){
					if(multiseq[j_eco]->seq[i1] == '|') printf("|\t-\t-");
					else printf("%c\t%ld\t-",multiseq[j_eco]->seq[i1],i3++);
				}
				i1++;
			}
		}
	}
	if(option == 2 || option == 3) printf("\n");
	
	for(j=0;j<num_seq;j++){
		multiseq[j]->seq[i1] = '\0';
		multiseq[j]->seqlen = i1;
	}

	return FALSE;
}


/***********************************************************/
int SelectSitesByInf(Seqfile **multiseq,double min, double max, enum data_type dtype1)
/* In a multiple alignment, sites that have information content
between min and max are selected. 
Gaps are not counted.*/
/***********************************************************/
{
	long i1,i2,j;
	double inf_content;
	double inf_contentAll;
	double inf_content_min,inf_content_max;
	long seqlen = multiseq[0]->seqlen;
	long seqlenAll = num_seq * seqlen;
	char *seqAll;
	char *seqTemp;

	if(min < 0.0 || max > 2.0){
		fprintf(stderr,"Unusual setting of min or max in SelectSitesByInf.\n");
		return TRUE;
	}

	if((seqAll=(char*)calloc(seqlenAll + 2,sizeof(char)))==NULL){
		fprintf(stderr,"Memory allocation error in SelectSitesByInf 1.\n");
		exit(1);
	}
	if((seqTemp=(char*)calloc(seqlen + 2,sizeof(char)))==NULL){
		fprintf(stderr,"Memory allocation error in SelectSitesByInf 1.\n");
		return -1.0;
	}

	for(j=0;j<num_seq;j++){
		strcat(seqAll,multiseq[j]->seq);
	}
	inf_contentAll = InfContent(seqAll,seqlenAll,dtype1);

	printf("Information content of whole alignment: %5.3e\n",inf_contentAll);

	inf_content_min = inf_contentAll * min;
	inf_content_max = inf_contentAll * max;

	for(i1=0,i2=0;i2<seqlen;i2++){
		for(j=0;j<num_seq;j++){
			seqTemp[j] = multiseq[j]->seq[i2];
		}
		seqTemp[num_seq] = '\0';
		inf_content = InfContent(seqTemp,num_seq,dtype1);
		if(inf_content_min <= inf_content && inf_content <= inf_content_max){
			for(j=0;j<num_seq;j++){
				multiseq[j]->seq[i1] = multiseq[j]->seq[i2];
			}
			i1++;
		}
	}
	for(j=0;j<num_seq;j++){
		multiseq[j]->seq[i1] = '\0';
		multiseq[j]->seqlen = i1;
	}

	return FALSE;
}

/***********************************************************/
double InfContent(char *seq,long seqlen,enum data_type dtype1)
/* This function returns information content of a sequence.
Returns negative value on error. */ 
/***********************************************************/
{
	double infcontent=0.0;
	long j;
	long code_count[30];
	double comp[30];
	long total;

	if(seq == NULL || seqlen <= 10){
		fprintf(stderr,"Strange data in InfContent.\n");
		return -1.0;
	}

	initialize_comp(code_count);

	if(c_comp_char(seq,seqlen,code_count)){
		fprintf(stderr,"Error in counting bases in InfContent.\n\n");
		return -1.0;
	}

	total = 0;
	for(j=1;j<30;j++) total += code_count[j];

	if(total == 0) total = 1;
	for(j=1;j<30;j++) comp[j] = (double)code_count[j]/(double)total;

#ifdef VERBOUS
	printf("comp[A] = %lf, seqlen = %ld, total = %ld\n",comp[A],seqlen,total);
#endif

	infcontent = 0.0;	
	for(j=1;j<30;j++){	/* unusual characters are not counted. */
		if(comp[j]==0.0) continue;
		infcontent += comp[j] * log(comp[j]) / log(2.0);
	}
	infcontent = 0.0 - infcontent;

#ifdef VERBOUS
	printf("Infcontent = %lf\n",infcontent);
#endif

	return infcontent;
}

/***********************************************************/
void initialize_comp2(long *code_count,unsigned dim)
/***********************************************************/
{
	unsigned s;

	for(s=0;s<dim;s++){
		code_count[s]=0L;
	}
	return;
}


/***********************************************************/
void initialize_comp(long code_count[30])
/***********************************************************/
{
	int s;

	for(s=0;s<30;s++){
		code_count[s]=0L;
	}
	return;
}

/**************************************************/
int codon_usage_comp(Seqfile *seq_in, long *code_count)
/**************************************************/
{
	long i;
	int s;
	char string[5]="   ";

	if(seq_in==NULL) return TRUE;
	if(seq_in->seqlen==0) return 2;
	if(!strcmp(seq_in->seq,"")) return 2;
	if(code_count==NULL) return 3;

        for(i=0;i<(seq_in->seqlen-2);i+=3){
		string[0]=seq_in->seq[i];
		string[1]=seq_in->seq[i+1L];
		string[2]=seq_in->seq[i+2L];
		string[3]='\0';

/*		strcpy(string,extract_string(seq_in->seq,i+1,i+3));
*/

		s=codon_usage(string);

/*		if(s==NULL) {
			fprintf(stderr,"Error in codon_usage_comp.\n");
			return TRUE;
		}
*/
		if(s >= 0 && s < 64) code_count[s] += 1L;
	}
	return FALSE;
}
/************************************************/


/**************************************************/
int c_comp(Seqfile *seq_in, long code_count[30])
/**************************************************/
{

	if(seq_in==NULL) return TRUE;
	if(seq_in->seqlen==0) return 2;
	if(!strcmp(seq_in->seq,"")) return 2;

	return c_comp_char(seq_in->seq,seq_in->seqlen,code_count);
}

/**************************************************/
int c_comp_char(char *seq,long seqlen,long code_count[30])
/**************************************************/
{
	long i;
	int s;

	if(seq==NULL) return TRUE;
	if(seqlen==0) return 2;
	if(!strcmp(seq,"")) return 2;

	for(i=0;i<seqlen;i++){
		if(islower(s=(seq[i]))) s=toupper(s);
		switch(s){
			case 'A':
				++code_count[A];
				break;
			case 'C':
				++code_count[C];
				break;
			case 'D':
				++code_count[D];
				break;
			case 'E':
				++code_count[E];
				break;
			case 'F':
				++code_count[F];
				break;
			case 'G':
				++code_count[G];
				break;
			case 'H':
				++code_count[H];
				break;
			case 'I':
				++code_count[I];
				break;
			case 'K':
				++code_count[K];
				break;
			case 'L':
				++code_count[L];
				break;
			case 'M':
				++code_count[M];
				break;
			case 'N':
				++code_count[N];
				break;
			case 'P':
				++code_count[P];
				break;
			case 'Q':
				++code_count[Q];
				break;
			case 'R':
				++code_count[R];
				break;
			case 'S':
				++code_count[S];
				break;
			case 'T':
				++code_count[T];
				break;
			case 'U':
				++code_count[U];
				break;
			case 'V':
				++code_count[V];
				break;
			case 'W':
				++code_count[W];
				break;
			case 'X':
				++code_count[X];
				break;
			case 'Y':
				++code_count[Y];
				break;
			default:
				if(isprint(s)){
					++code_count[0];        /* zero stands for other characters. */
				}
				break;
		}
	}
	return FALSE;
}
/************************************************/


/******************************************************************************/
int MatchSeqN(char **seq1,char **seq2,char *seq_out,long seqlen_out,long match_len,double c,char *codon_table)
/*******************************************************************************
This function returns the position in seq2 (DNA) from which matching
with seq1 (protein) begins. The matching is judged with allowance c.
At the time of return, new values of seq1 and seq2 are also returned.
The aligned sequence is appended to seq_out.
Each of the new values point to the position from which unmatching of
sequences begins. This function returns 0, if no match is found.
match_len is normally 10, c is 0.8.
********************************************************************************/
{
	int i,j;
	int k;
	long seqlen1;	/* length of seq1 */
	long seqlen1c;	/* length of seq1 without hyphen */
	long seqlen2;	/* length of seq2 */
	long seqlen3;	/* length of translated seq2 */
	char *seqtr[3];	/* translated sequences */
	char seq3[100];	/* partial extracted sequence of translated seq2 */
	char *p,*q,*q0,*r,*pp;
	int count;
	long position;
	long min_position;
	int min_k = 0;
	int min_match;
	char *temp,*temp0;
	char *seq1c;	/* for seq1 without hyphen */

	if(seq1==NULL || seq2==NULL || seq_out==NULL || c<0.5 || c>1.0 ) return 0;
	seqlen1=strlen(*seq1);
	seqlen2=strlen(*seq2);
	if((seq1c=(char*)calloc(seqlen1 + 2,sizeof(char)))==NULL){
		fprintf(stderr,"Out of memory in MatchSeqN 1.\n");
#include "exit1.h"
	}
	strcpy(seq1c,*seq1);
	if(StringClearGaps(seq1c)){
		fprintf(stderr,"Error in clearing gaps in MatchSeqN.\n");
#include	"exit1.h"
	}
	seqlen1c = strlen(seq1c);	
	if(seqlen2 < seqlen1c * 3) return 0;
	if(match_len < 8) return 0;
	min_match = (int)((double)match_len * c);

	for(k=0;k<3;k++){
		if((seqtr[k]=translate(*seq2,k+1,codon_table))==NULL){
			fprintf(stderr,"MatchSeqN: Unable to translate sequence in frame %d.\n",k+1);
#include "exit1.h"
		}
	}

/* for debug printf("1: %s\n2: %s\n3: %s\n",seqtr[0],seqtr[1],seqtr[2]); */ 
	if((temp0=(char*)calloc(seqlen_out + 2,sizeof(char)))==NULL){
		fprintf(stderr,"Out of memory in MatchSeqN 2.\n");
#include "exit1.h"
	}
	temp = temp0;
	position = min_position = seqlen3 = seqlen2 / 3 + 1;
	if(strlen(seq1c) < match_len) return 0;

	strncpy(seq3,seq1c,match_len);
	seq3[match_len] = '\0';

#ifdef VERBOUS
	printf("seq3= %s, match_len= %ld\n",seq3,match_len);
	fflush(stdout);
#endif

	for(k=0;k<3;k++){
		position = MatchSeq(seq3,seqtr[k],c);

#ifdef VERBOUS
	printf("k=%d, position=%ld\n",k,position);
#endif
		if(position != 0 && position < min_position){
			min_position = position;
			min_k = k;
		}
	}
	
	if(**seq1 == '-'){
		for(p=*seq1;*p=='-';p++){
			*temp++ = '-';
			*temp++ = '-';
			*temp++ = '-';
		}
		*seq1 = p;
	}
		
	q0 = seqtr[min_k]+(min_position-1);	/* pointer to the translated sequence */
	r = *seq2 + (min_k + 3 * (min_position - 1));	/* pointer to the DNA sequence */
	for(p=*seq1,q=q0;*p!='\0'&& *q!='\0';p++){
		if(*p == '-'){
			*temp++ = '-';
			*temp++ = '-';
			*temp++ = '-';
			continue;
		} else if(*p == *q){
			q++;
			*temp++ = *r++;
			*temp++ = *r++;
			*temp++ = *r++;
			continue;
		} else if(*q == '*'){
			for(i=0;i<6;i++){
				if( *(p+i) == '\0' ) break;
			}
			if(i==6){
				q++;
				*temp++ = *r++;
				*temp++ = *r++;
				*temp++ = *r++;
				continue;
			} else {
				while(i>0){
					*temp++ = '-';
					*temp++ = '-';
					*temp++ = '-';
					i--;
				}
				break;
			}
		} else {
			/* detection of 5' continuous gaps */
			count = 0;
			for(i=1;i<(2 * match_len);i++){
				if(*(p+i) == '\0') break;
				else if(*(p+i) == '-') {
					count++;
					continue;
				} else {
					count=0;
				}
			}
			if(*(p+i)=='\0' && count > 0){
				while(*p!='\0'){
					if(*p++ == '-'){
						*temp++ = '-';
						*temp++ = '-';
						*temp++ = '-';
					} else {
						q++;
						*temp++ = tolower(*r++);
						*temp++ = tolower(*r++);
						*temp++ = tolower(*r++);
					}
				}
				break;
			}
			
			count = 0;
			for(i=0,j=0;i<match_len;i++){
				if(*(p+(i+j)) == '-'){
					j++;
					i--;
					continue;
				} else if(*(p+(i+j)) == '\0' || *(q+i) == '\0'){
					count = 0;
					break;
				} else if(*(p+(i+j)) == *(q+i)){
					count++;
					continue;
				}
			}
			if(i<=1 && j>0){
				while(j>0){
					*temp++ = '-';
					*temp++ = '-';
					*temp++ = '-';
					j--;
				}
				break;
			} else if(i < match_len){
				while(i > 0){
					if(*p++ == '-'){
						*temp++ = '-';
						*temp++ = '-';
						*temp++ = '-';
					} else {
						q++;
						*temp++ = tolower(*r++);
						*temp++ = tolower(*r++);
						*temp++ = tolower(*r++);
					}
					i--;
				}
				break;
			} else if(count >= min_match){
				q++;
				*temp++ = tolower(*r++);
				*temp++ = tolower(*r++);
				*temp++ = tolower(*r++);
				continue;
			} else {
				break;
			}
		}
	}

	*temp = '\0';
	if((strlen(seq_out) + strlen(temp0)) <= seqlen_out){
		strcat(seq_out,temp0);
	} else {
		printf("Length of seq_out is %ld, length of temp0 is %ld, seqlen1 = %ld.\n",strlen(seq_out),strlen(temp0),seqlen_out);
	}

#ifdef VERBOUS
	printf("seq1=%ld\n seq2=%ld\n\np=%ld\n",*seq1,*seq2,p);
	printf("temp0= %s\nseq_out= %s\n",temp0,seq_out);
	fflush(stdout);
#endif

	/* Both seq1 and seq2 points to the positions from which sequences do not match. */
	*seq1 = p;
	*seq2 += (min_k + 3 * ((q - q0)/sizeof(char)));

	/* check if all characters in seq1 are gaps */
	for(pp = *seq1; *pp != '\0'; pp++){
		if(*pp != '-') break;
	}

/*	*seq2 = r;
*/
	if(*p=='\0') return 0;
	else if(*pp=='\0') return 0;
	else return 1;
}
/*********************************************/


/*********************************************/
long MatchSeq(char *seq1,char *seq2,double c)
/*********************************************
This function returns the position in seq2 
from which matching with seq1 begins.
The matching is judged with allowance c.
**********************************************/
{
	long i,j,k1,k2;
	long seqlen1;	/* length of seq1 */
	long seqlen2;	/* length of seq2 */
	long diff;	/* correction for stop codons in seq2 */
	long match;	/* number of matched residues */
	long match1;	/* local matches */
	long match2;	/* local matches */

	double c0;	/* ratio of matched residues */

	if(seq1==NULL || seq2==NULL || c<0.5 || c>1.0 ) return 0L;
	seqlen1=strlen(seq1);
	seqlen2=strlen(seq2);
	if(seqlen2<seqlen1) return 0L;

	for(i=0L;i<seqlen2-seqlen1+1;i++){
		diff=0L;
		match=0L;
		for(j=0;j<seqlen1;j++){	
			if(!isalpha(seq2[j+i+diff])) {
				match1=0L;
				for(k1=j+1,k2=j+i+diff+1;k1<j+6;k1++,k2++){
					if(seq1[k1]==seq2[k2]) match1+=1L;
				}
				match2=0L;
				for(k1=j+1,k2=j+i+diff+2;k1<j+6;k1++,k2++){
					if(seq1[k1]==seq2[k2]) match2+=1L;
				}
				if(match1<4 && match2>match1) {
					diff+=1L;
					continue;
				}
			}
			if(seq1[j]==seq2[j+i+diff]) match+=1L;	
		}

/* for debug 
	printf("i=%ld,j=%ld,match=%ld,seqlen1=%ld,seqlen2=%ld\n",i,j,match,seqlen1,seqlen2); 
*/
		c0=(double)match/(double)seqlen1;
		if(c0>c) break;
	}
	if(i>=seqlen2-seqlen1+1) return 0;
	
	return i+1;
}
/*********************************************/


long num_seq_nuc;

/******************************************************************/
int massAlnCorrect(char *fnameList,char *fname2,double thr,char *nuc_file,long offset)
/******************************************************************/
{
	FILE *fin;
	long i;
	int k;
	int counter;
	char str1[MAXLINE + 2];
	char fname1[100];
	int c=0;
	Seqfile **seq_nuc,**seqM;
	enum data_type dtype1=unknown;
	enum data_type dtype2=unknown;
//	long num_seq_nuc;
	long num_seq_bak;
	char nuc_file_new[100];
	int _max_aln = 2000;	/* Maximum limit of number of sequences in an alignment */
	int _max_length = 10000;	/* Maximum length of an alignment */

	strcpy(nuc_file_new,nuc_file);
	strcat(nuc_file_new,".rev");

	/* nucleic acid sequence is read at first */
	fprintf(stdout,"Reading nucleic acid sequences...\n");
	num_seq_bak=num_seq;
	if((seq_nuc=Mread_fasta(nuc_file,dtype2))==NULL){
		fprintf(stderr,"Error in reading file: %s\n",nuc_file);
		return TRUE;
	}
	num_seq_nuc=num_seq;
	dtype2=check_data_type(seq_nuc[0]);
	if(dtype2!=DNA){
		fprintf(stderr,"These are not DNA sequences.\n");
		return TRUE;
	}
	num_seq = num_seq_bak;

	/* reading a list of alignments */
	fprintf(stdout,"Reading list file ...\n");
	if((fin=fopen(fnameList,"r"))==NULL){
		fprintf(stderr,"Cannot open file: %s\n\n",fnameList);
		return TRUE;
	}

	/* allocating memory to seqM */
	if((seqM=(Seqfile**)calloc(1,_max_aln*sizeof(Seqfile)))==0){
		fprintf(stderr,"Out of memory in mass_alnCorr 1\n");
		return TRUE;
	}
	for(k=0;k<_max_aln;k++){
		if((seqM[k]=(Seqfile *)calloc(1,sizeof(Seqfile)))==0){
			fprintf(stderr,"Out of memory in mass_alnCorr 2\n");
			return TRUE;
		}
		if((seqM[k]->ID=(char *)calloc(MINMEMSIZE,sizeof(char)))==0){
			fprintf(stderr,"Out of memory in mass_alnCorr 3\n");
			return TRUE;
		}
		if((seqM[k]->header=(char *)calloc(MINMEMSIZE,sizeof(char)))==0){
			fprintf(stderr,"Out of memory in mass_alnCorr 4\n");
			return TRUE;
		}
		if((seqM[k]->seq=(char *)calloc(_max_length,sizeof(char)))==0){		
			fprintf(stderr,"Out of memory in mass_alnCorr 5\n");
			return TRUE;
		}
	}
	
	counter = 0;
/* A loop of reading fin should embrace all these until the end*/	
	while(fgets(str1,MAXLINE+1,fin)!=NULL){
		if(counter == 0 && *str1 == '>'){
			counter = 1;
			continue;
		}
		if(counter == 0) continue;	
		if(*str1 == '#') continue;
		counter += 1;
		sscanf(str1,"%s",fname1);
		c=check_filetype(fname1);
		if(counter > 1000){
			printf("Now 1000 files have been processed.\n");
			printf("Please edit the list file to indicate the position to start\n");
			printf("by inserting a > mark before the line: %s.\n",fname1);
			break;
		}
		if(c==CLUSTAL_SEQ){
#if 0
			if((seqM=read_clustal(fname1,dtype1))==NULL){
				fprintf(stderr,"Error in reading file: %s\n",fname1);
				continue;
			}
#endif
		/* new */
			read_clustal2(fname1,seqM);
			dtype1=check_data_type(seqM[0]);

		}else {
			fprintf(stderr,"This is not a CLUSTAL file: %s\n",fname1);
			continue;
		}
#if 0
		for(i=0;i<num_seq;i++){
			if((seqM[i]->header = (char*)calloc(10, sizeof(char)))==NULL){
				fprintf(stderr,"Memory allocation error in mass_alnCorrect.\n\n");
			}
		}
#endif
		if(alnCorrect(2,fname1,fname2,seqM,thr,nuc_file,seq_nuc,"all",offset)){
			fprintf(stderr,"Error in mass_alnCorrect mode 2.\n");
			return TRUE;
		}
	}	/* end of while */
	fclose(fin);
	
	for(i=0;i<num_seq_nuc;i++){
		if(write_fasta(nuc_file_new,seq_nuc[i],1,' ')!=0){
			fprintf(stderr,"Error in writing a new nucleotide sequence file.\n");
			return TRUE;
		}
	}

	return FALSE;
}
/******************************************************************/



static char gap_space[20000];	/* In place of space_line function */
/******************************************************************/
void clearGapSpace(void){
	int i;
	for(i=0;i<19999;i++) gap_space[i] = '-';
	gap_space[19999] = '\0';
}
/******************************************************************/

/******************************************************************/
void GapSpace(int j){
	int i;
	for(i=0;i<19999;i++) gap_space[i] = '-';
	gap_space[j] = '\0';
}
/******************************************************************/

/******************************************************************/
int alnCorrect(int mode,char *fname1,char *fname2,Seqfile **seqM,double thr,char *nuc_file,Seqfile **seq_nuc,char *org_name,long offset)
/* fname1 is an aln file. */
/* mode 1: single mode. mode 2: multiple mode. */
/******************************************************************/
{
	FILE *fout;
	long i,j,k,k0,k1,k2,kk,ii;
	long length;
	double total;
	double total2;
	double stdev;
	double av1,av2,av3,std1,std2,std3,av1x,std1x;
	double num_seq_thr;
	long seqlen = seqM[0]->seqlen;
	long num_seq_local,seqlen_local;
//	Seqfile **seq_nuc;
	enum data_type dtype2=unknown;
//	long num_seq_nuc;
	long num_seq_bak;
	long num_seq_new;
	static char ID[100];
	long offset3 = offset / 3;
	long offset2 = offset - 3 * offset3;
	long offset4 = 3 * offset3;
#if 1
	double *mat;
	int mat_size = 15 * num_seq + 200;
	char seq_temp[offset + 3 * seqlen + 1000];
	char seq_aa0[offset3 + 2 * seqlen + 1000];
	char seq_aa1[offset3 + 2 * seqlen + 1000];
	char seq_aa2[offset3 + 2 * seqlen + 1000];
	char seq_aa3[offset3 + 2 * seqlen + 1000];
#endif
/* These could be defined as static to reduce memory usage */
/* In this case, memory allocation executed below should be commented out */
#if 0	
	int mat_size = 20000;
	double mat[20000];
	static char seq_temp[3000];
	static char seq_aa0[1000];
	static char seq_aa1[1000];
	static char seq_aa2[1000];
	static char seq_aa3[1000];
#endif
//	static char gap_space[20000];	/* In place of space_line function */
	char memo[MAXLINE+MINMEMSIZE];
	long position;
	static char codon_table[100];
	long diff2,diff_position,av1_position,new_position,Nterm_size;
	char newname[100];
	char *x;
	Boolean flag=FALSE;
	Boolean print_flag = FALSE;
	Boolean low_flag,high_flag;
	Boolean short_sequence=FALSE;
	Boolean no_DNA = FALSE;
	long gaps;
	long Nterm_size3,Nterm_size0,k0m,k0v,k0l;
	char fname2_org[100];
	long new_size=100L;
/*	long temp_size=100L;	*/
	char *np;
	long k3k0;
	long slide,seqlen2;

	strcpy(fname2_org,fname2);
	strcpy(newname,fname1);
	x = strstr(newname,".aln");
	*x = '\0';
	strcat(newname,".rev.aln");
	strcpy(codon_table,CodonTable("code101"));
	clearGapSpace();

/* mat_size is fixed: 20000 */
#if 1
	if((mat = (double*)calloc(mat_size, sizeof(double)))==NULL){
		fprintf(stderr,"Memory allocation error in alnCorrect.\n\n");
		return TRUE;
	}
#endif
	for(i=0;i<mat_size;i++) mat[i] = 0.0;

	if(num_seq > 1000){
		printf("Number of sequences should be at most 1000. %s %ld\n",fname1,num_seq);
		return FALSE;
	}
	strcpy(ID,"");
	
	/* nucleic acid sequence is read at first in correcting all sequences */
	if(mode==1 && strcmp(org_name,"all")==0){
		fprintf(stdout,"Reading nucleic acid sequences...\n");
		num_seq_bak=num_seq;
		if((seq_nuc=Mread_fasta(nuc_file,dtype2))==NULL){
			fprintf(stderr,"Error in reading file: %s\n",nuc_file);
			return TRUE;
		}
		num_seq_nuc=num_seq;
		dtype2=check_data_type(seq_nuc[0]);
		if(dtype2!=DNA){
			fprintf(stderr,"These are not DNA sequences.\n");
			return TRUE;
		}
		num_seq = num_seq_bak;
	}

#if 1
	/* Detecting short sequences */
	total = 0.0;
	total2 = 0.0;
	for(i=0;i<num_seq;i++){
		length = 0L;
		for(j=0;j<seqlen;j++){
			if(seqM[i]->seq[j] != '-') length += 1L;
		}
		mat[num_seq * 4 + i] = (double)length;
		total +=  (double)length;
		total2 += (double)(length * length);
	}
	total /= (double)num_seq;
	stdev = (total2 / (double)num_seq) - total * total;
	stdev = sqrt(stdev);
	av3 = total;
	std3 = stdev;
	/* Removing short sequences if any */
	flag = FALSE;
	num_seq_new = num_seq;
	for(i=0;i<num_seq_new;i++){
		mat[num_seq * 5 + i] = (mat[num_seq * 4 + i] - av3 ) / std3;
		if((mat[num_seq * 5 + i] < 2.0) && (mat[num_seq * 4 + i] * 2.0 < av3)){
			printf("Short sequence: %s length=%ld ",seqM[i]->ID, (long)mat[num_seq * 4 + i]);
			fprintf(stdout,"is being removed from the alignment\n");
			fflush(stdout);
			if(i == num_seq_new -1){
				num_seq_new -= 1L;
				break;
			}else{
				for(ii=i; ii<num_seq_new-1;ii++){
					for(k=0;k<15;k++) mat[num_seq * k + ii] = mat[num_seq * k + ii + 1];
					strcpy(seqM[ii]->ID,seqM[ii+1]->ID);
					strcpy(seqM[ii]->header,seqM[ii+1]->header);
					strcpy(seqM[ii]->seq,seqM[ii+1]->seq);
				}
				flag = TRUE;
				num_seq_new -= 1L;
				i--;
			}
		}
	}
	
	if(flag){
		num_seq = num_seq_new;
		for(i=0;i<mat_size;i++) mat[i] = 0.0;
	}

	/* End of short sequences */
#endif

/* Start with short sequences removed */
	total = 0.0;
	total2 = 0.0;
	for(i=0;i<num_seq;i++){
		for(j=0;j<seqlen;j++){
			if(seqM[i]->seq[j] != '-') break;
		}
		mat[num_seq * 0 + i] = (double)j;
		total +=  (double)j;
		total2 += (double)(j * j);
	}
	total /= (double)num_seq;
	stdev = (total2 / (double)num_seq) - total * total;
	stdev = sqrt(stdev);
	av1 = total;
	std1 = stdev;


	for(i=0;i<num_seq;i++){
		if(std1 < 0.1) mat[num_seq * 1 + i] = 0.01;
		else mat[num_seq * 1 + i] = (mat[num_seq * 0 + i] - av1 ) / std1;
		num_seq_local = 0L;
		seqlen_local = mat[num_seq * 0 + i];
		total = total2 = stdev = 0.0;
		for(j=0;j<num_seq;j++){
			if(fabs(mat[num_seq * 0 + j] - seqlen_local) <= std1){
				num_seq_local += 1L;
				total += (double)mat[num_seq * 0 + j];
				total2 += (double)(mat[num_seq * 0 + j] * mat[num_seq * 0 + j]);
			}
		}
		total /= (double)num_seq_local;
		stdev = (total2 / (double)num_seq_local) - total * total;
		stdev = sqrt(stdev);
		mat[num_seq * 6 + i] = (double)num_seq_local;
		mat[num_seq * 7 + i] = total;
		mat[num_seq * 8 + i] = stdev;
	}

	/* excluding outliers */
	num_seq_local = 0L;
	total = total2 = stdev = 0.0;
	for(j=0;j<num_seq;j++){
		if(fabs(mat[num_seq * 0 + j] - av1) <= std1){
			num_seq_local += 1L;
			total += (double)mat[num_seq * 0 + j];
			total2 += (double)(mat[num_seq * 0 + j] * mat[num_seq * 0 + j]);
		}
	}
	total /= (double)num_seq_local;
	stdev = (total2 / (double)num_seq_local) - total * total;
	stdev = sqrt(stdev);
	av1x = total;
	std1x = stdev;



	total = 0.0;
	total2 = 0.0;
	for(i=0;i<num_seq;i++){
		for(j=0;j<seqlen;j++){
			if(seqM[i]->seq[seqlen-j-1] != '-') break;
		}
		mat[num_seq * 2 + i] = (double)j;
		total +=  (double)j;
		total2 += (double)(j * j);
	}
	total /= (double)num_seq;
	stdev = (total2 / (double)num_seq) - total * total;
	stdev = sqrt(stdev);
	av2 = total;
	std2 = stdev;

	for(i=0;i<num_seq;i++){
		if(std2 < 0.1) mat[num_seq * 3 + i] = 0.01;
		else mat[num_seq * 3 + i] = (mat[num_seq * 2 + i] - av2) / std2;
		num_seq_local = 0L;
		seqlen_local = mat[num_seq * 2 + i];
		total = total2 = stdev = 0.0;
		for(j=0;j<num_seq;j++){
			if(fabs(mat[num_seq * 2 + j] - seqlen_local) <= std2){
				num_seq_local += 1L;
				total += (double)mat[num_seq * 2 + j];
				total2 += (double)(mat[num_seq * 2 + j] * mat[num_seq * 2 + j]);
			}
		}
		total /= (double)num_seq_local;
		stdev = (total2 / (double)num_seq_local) - total * total;
		stdev = sqrt(stdev);
		mat[num_seq * 9 + i] = (double)num_seq_local;
		mat[num_seq * 10 + i] = total;
		mat[num_seq * 11 + i] = stdev;
	}


	total = 0.0;
	total2 = 0.0;
	for(i=0;i<num_seq;i++){
		length = 0L;
		for(j=0;j<seqlen;j++){
			if(seqM[i]->seq[j] != '-') length += 1L;
		}
		mat[num_seq * 4 + i] = (double)length;
		total +=  (double)length;
		total2 += (double)(length * length);
	}
	total /= (double)num_seq;
	stdev = (total2 / (double)num_seq) - total * total;
	stdev = sqrt(stdev);
	av3 = total;
	std3 = stdev;

	for(i=0;i<num_seq;i++){
		if(std3 < 0.1) mat[num_seq * 5 + i] = 0.01;
		else mat[num_seq * 5 + i] = (mat[num_seq * 4 + i] - av3 ) / std3;
		num_seq_local = 0L;
		seqlen_local = mat[num_seq * 4 + i];
		total = total2 = stdev = 0.0;
		for(j=0;j<num_seq;j++){
			if(fabs(mat[num_seq * 4 + j] - seqlen_local) <= std3){
				num_seq_local += 1L;
				total += (double)mat[num_seq * 4 + j];
				total2 += (double)(mat[num_seq * 4 + j] * mat[num_seq * 4 + j]);
			}
		}
		total /= (double)num_seq_local;
		stdev = (total2 / (double)num_seq_local) - total * total;
		stdev = sqrt(stdev);
		mat[num_seq * 12 + i] = (double)num_seq_local;
		mat[num_seq * 13 + i] = total;
		mat[num_seq * 14 + i] = stdev;
	}

	/* end of statistics */



	if(mode==2){
		strcpy(fname2,fname2_org);
		strcat(fname2,".");
		strcat(fname2,fname1);
	}	
	print_flag = FALSE;
	if(strcmp(org_name,"all")!=0) strcat(org_name,"_");
	num_seq_thr = (double)num_seq * thr;
	num_seq_new = num_seq;

	for(i=0;i<num_seq_new;i++){
		if(strcmp(org_name,"all")!=0){
			if(!linetype(seqM[i]->ID,org_name)) continue;
		}


if(mat[num_seq * 0 + i] < 0.0 || mat[num_seq * 0 + i] > (double)seqlen){
	fprintf(stderr,"Site2: Aln: %s: mat[0] is strange for i=%ld: %s\n",fname1,i,seqM[i]->ID);
}


		if(print_flag==FALSE){
			if((fout=fopen(fname2,"w"))==NULL){
				fprintf(stderr,"Cannot open file: %s\n\n",fname2);
				return TRUE;
			}
			strcpy(fname2,fname2_org);
			print_flag=TRUE;
			fprintf(fout," Statistics of alignment: %s. Output of SISEQ\n\n",fname1);
			fprintf(fout," Number of sequences: %ld.\n",num_seq);
			fprintf(fout," Start position: average = %6.3f stdev = %6.3f.\n",av1,std1);
			fprintf(fout,"   without outliers: average = %6.3f stdev = %6.3f.\n",av1x,std1x);
			fprintf(fout," C-terminal vacant space: average = %6.3f stdev = %6.3f.\n",av2,std2);
			fprintf(fout," Sequence length: average = %6.3f stdev = %6.3f.\n",av3,std3);
			fprintf(fout,"\n ID\tstart_av\tstart_dev\tend_av\tend_dev\tlength_av\tlength_dev");
			fprintf(fout,"\tStart_neighbor #seq\taverage\tstdev");
			fprintf(fout,"\tEnd_neighbor #seq\taverage\tstdev");
			fprintf(fout,"\tLength_neighbor #seq\taverage\tstdev");
			fprintf(fout,"\n");
		}


		if(mat[num_seq * 6 + i] >= num_seq_thr && mat[num_seq * 9 + i] >= num_seq_thr && mat[num_seq * 12 + i] >= num_seq_thr){
			continue;
		}

		printf("Processing sequence %s in the alignment: %s\n",seqM[i]->ID,fname1);
		short_sequence = FALSE;
		no_DNA = FALSE;

		if(mode==1 && strcmp(org_name,"all")!=0){
		/* nucleic acid sequence is read only if necessary */
			fprintf(stdout,"Reading nucleic acid sequences...\n");
			num_seq_bak=num_seq;
			if((seq_nuc=Mread_fasta(nuc_file,dtype2))==NULL){
				fprintf(stderr,"Error in reading file: %s\n",nuc_file);
				return TRUE;
			}
			num_seq_nuc=num_seq;
			dtype2=check_data_type(seq_nuc[0]);
			if(dtype2!=DNA){
				fprintf(stderr,"These are not DNA sequences.\n");
				return TRUE;
			}
			num_seq = num_seq_bak;
		}


/* mat is sometimes strange */
#if 0
		if(mat[num_seq * 0 + i] < 0.0 || mat[num_seq * 0 + i] > (double)seqlen){
			fprintf(stderr,"Site0: Aln: %s: mat[0] is strange for i=%ld: %s\n",fname1,i,seqM[i]->ID);
			for(j = 0;j < 15;j++){
				if(j == 6 || j == 9 || j == 12){
					fprintf(stderr,"\t%ld",(long)mat[num_seq * j + i]);
				} else {
					fprintf(stderr,"\t%6.3f",mat[num_seq * j + i]);
				}
			}
			fprintf(stderr,"\n");
			fflush(stderr);

			for(j=0;j<seqlen;j++){
				if(seqM[i]->seq[j] != '-') break;
			}
			mat[num_seq * 0 + i] = (double)j;
			mat[num_seq * 1 + i] = (mat[num_seq * 0 + i] - av1 ) / std1;
		}else if(mat[num_seq * 2 + i] < 0.0 || mat[num_seq * 2 + i] > (double)seqlen){
			fprintf(stderr,"mat[2] is strange for i=%ld\n",i);
			for(j=0;j<seqlen;j++){
				if(seqM[i]->seq[seqlen-j-1] != '-') break;
			}
			mat[num_seq * 2 + i] = (double)j;
		}
#endif

		if(print_flag) fprintf(fout,"%s",seqM[i]->ID);
		for(j = 0;j < 15;j++){
			if(j == 6 || j == 9 || j == 12){
				if(print_flag) fprintf(fout,"\t%ld",(long)mat[num_seq * j + i]);
			} else {
				if(print_flag) fprintf(fout,"\t%6.3f",mat[num_seq * j + i]);
			}
		}
		if(print_flag) fprintf(fout,"\n");
		fflush(fout);

	/* Correct option */
		/* short sequences are skipped. */
		if((mat[num_seq * 5 + i] < 2.0) && (mat[num_seq * 4 + i] * 2.0 < av3)){
			short_sequence = TRUE;
			printf("Short sequence: No.=%ld %s length=%ld.\n",i,seqM[i]->ID, (long)mat[num_seq * 4 + i]);
		}

		if(std1x < 1.1) std1x = 1.1;
		av1_position = (long)av1x;
		if(!short_sequence && (mat[num_seq * 6 + i] < num_seq_thr)){
			strcpy(ID,seqM[i]->ID);
			position = (long)mat[num_seq * 0 + i];
			diff_position = position - av1_position;

			/* Protruding N-terminus */
			if(diff_position < 0L){
				Nterm_size = 5L - diff_position + (long)std1;
				Nterm_size0 = 1L - diff_position;
				if(Nterm_size > offset) Nterm_size = offset;
				strncpy(seq_aa0,seqM[i]->seq+position,Nterm_size);
				seq_aa0[Nterm_size] = '\0';
				if(print_flag) fprintf(fout,"N-terminus is protruding.\n");
				GapSpace(Nterm_size - 5);
				if(print_flag) fprintf(fout,"%s%s\n",gap_space,"v");
//				if(print_flag) fprintf(fout,"%s%s\n",space_line(Nterm_size - 5,'-'),"v");
				if(print_flag) fprintf(fout,"%s\n",seq_aa0);
				fflush(fout);
if(mat[num_seq * 0 + i] < 0.0 || mat[num_seq * 0 + i] > (double)seqlen){
	fprintf(stderr,"Site7a: Aln: %s: mat[0] is strange for i=%ld: %s\n",fname1,i,seqM[i]->ID);
}
if(mat[num_seq * 0 + i+1] < 0.0 || mat[num_seq * 0 + i+1] > (double)seqlen){
	fprintf(stderr,"Site7b: Aln: %s: mat[0] is strange for i=%ld: %s\n",fname1,i+1,seqM[i+1]->ID);
}

				for(j=0;j<num_seq_nuc;j++){
					if(!strcmp(ID,seq_nuc[j]->ID)){
						Nterm_size3 = 3 * Nterm_size;
						strncpy(seq_temp,seq_nuc[j]->seq+offset,Nterm_size3);
						seq_temp[Nterm_size3] = '\0';
						strcpy(seq_aa1,translate(seq_temp,1,codon_table));

						/* Gaps are inserted */
						for(kk=0,k=0;kk<Nterm_size;kk++){
							if(k == 0){
								if(seq_aa1[k] == 'v' || seq_aa1[k] == 'l'){
									seq_aa1[k] = 'M';
								}
							}
							if(seq_aa0[kk] == '-'){
								seq_aa2[kk] = '-';
							}else if(seq_aa0[kk] == toupper(seq_aa1[k])){
								seq_aa2[kk] = seq_aa1[k++];
							}else{
								seq_aa2[kk] = '?';
								k++;
							}
						}
						strcpy(seq_aa1,seq_aa2);
						strcpy(seq_aa2,"");
						if(print_flag) fprintf(fout,"%s\n\n",seq_aa1);
						fflush(fout);

if(mat[num_seq * 0 + i] < 0.0 || mat[num_seq * 0 + i] > (double)seqlen){
	fprintf(stderr,"Site6a: Aln: %s: mat[0] is strange for i=%ld: %s\n",fname1,i,seqM[i]->ID);
}
if(mat[num_seq * 0 + i+1] < 0.0 || mat[num_seq * 0 + i+1] > (double)seqlen){
	fprintf(stderr,"Site6b: Aln: %s: mat[0] is strange for i=%ld: %s\n",fname1,i+1,seqM[i+1]->ID);
}
						/* Putative initiation codon within the ORF should be searched based on DNA sequence. */
						k0 = 0;
						k0m = k0v = k0l = 0L;
						for(k = Nterm_size0;k > 0 ;k--){
							if(seq_aa1[k] == 'm'){
								k0m = k;
								break;
							}
						}
						for(k = Nterm_size0;k < Nterm_size ;k++){
							if(seq_aa1[k] == 'm'){
								if(k + k0m < 2 * Nterm_size0) k0m = k;
								break;
							}
						}
						for(k = Nterm_size0;k > 0 ;k--){
							if(seq_aa1[k] == 'v'){
								k0v = k;
								break;
							}
						}
						for(k = Nterm_size0;k < Nterm_size ;k++){
							if(seq_aa1[k] == 'v'){
								if(k + k0v < 2 * Nterm_size0) k0v = k;
								break;
							}
						}
						for(k = Nterm_size0;k > 0 ;k--){
							if(seq_aa1[k] == 'l'){
								k0l = k;
								break;
							}
						}
						for(k = Nterm_size0;k < Nterm_size;k++){
							if(seq_aa1[k] == 'l'){
								if(k + k0l < 2 * Nterm_size0) k0l = k;
								break;
							}
						}
#if 0
fprintf(fout,"k0m=%ld, k0v=%ld, k0l=%ld: Nterm_size0=%ld, Nterm_size=%ld, std1=%ld\n",k0m,k0v,k0l,Nterm_size0,Nterm_size,(long)std1);
fflush(fout);
#endif
						if(k0m > Nterm_size0 - (long)std1 - 1) k0 = k0m;
						else if(k0v > Nterm_size0 - (long)std1 - 1) k0 = k0v;
						else if(k0l > Nterm_size0 - (long)std1 - 1) k0 = k0l;
						if(k0 == 0){
							if(k0m > 0) k0 = k0m;
							else if(k0v > 0) k0 = k0v;
							else if(k0l > 0) k0 = k0l;
						}
						if(k0 > 0){
							k2 = 0;
							if(print_flag) fwrite_fasta(fout,seqM[i],0,' ');
							kk = position - 1L;
							if(kk < 0L) kk = 0L;
							for(k=0L;k < k0;k++){
								if(k < 0L || kk >= seqlen) break;	/* check */
								if((seqM[i]->seq)[kk] == toupper(seq_aa1[k])){
									(seqM[i]->seq)[kk++] = '-';
									k2 += 1;
								}else if(seq_aa1[k] == '?'){
									(seqM[i]->seq)[kk++] = '-'; 
									k2 += 1;
								}else {
									k -= 1L;
									kk += 1L;
								}
							}
							strcpy(seqM[i]->header,"revised");

							if(print_flag) fwrite_fasta(fout,seqM[i],1,' ');
							if(print_flag) fprintf(fout,"\n");
							fflush(fout);
							/* modifying nuc sequence for protruding N-term */
							k3k0 = 3 * k2;	/* k2 is the number of real aa */
							for(np=seq_nuc[j]->seq;*(np+k3k0)!='\0';np++){
								*np = *(np + k3k0);
							}
							*np = '\0';
							seq_nuc[j]->seqlen = strlen(seq_nuc[j]->seq);
							if(strstr(seq_nuc[j]->header,"Revised:")!=NULL && strlen(seq_nuc[j]->header) < MAXLINE ){
								sprintf(memo,": %ld bases removed from 5' terminus",k3k0);
								strcat(seq_nuc[j]->header,memo);
							}else{
								sprintf(memo,"Revised: %ld bases removed from 5' terminus",k3k0);
								strcpy(seq_nuc[j]->header,memo);
							}
							/* end of modifying nuc */

							flag = TRUE;
						}
						break;
					}
				}
				if(j == num_seq_nuc){
					fprintf(fout,"DNA sequence was not found for %s.\n",ID);
					no_DNA = TRUE;
				}
			/* end of protruding N-term */




			} else {
			/* N-term extention */	
/*********** start shifting of code ******/
			/* counting gaps */
			gaps = 0L;
			for(j=av1_position;j<position;j++){
				if(j < 0) continue;
				for(k=0;k<num_seq;k++){
					if(k == i) continue;
					if((seqM[k]->seq)[j] == '-') gaps += 1;
				}
			}
			if(num_seq == 1L){
				gaps = 0L;
			} else gaps /= (num_seq - 1L);

			for(j=0;j<num_seq_nuc;j++){
				if(!strcmp(ID,seq_nuc[j]->ID)){
					strncpy(seq_temp,(seq_nuc[j]->seq)+offset2,offset4);
					seq_temp[offset4] = '\0';
					strcpy(seq_aa1,translate(seq_temp,1,codon_table));
					strcpy(seq_aa2,translate(seq_temp,2,codon_table));
					strcpy(seq_aa3,translate(seq_temp,3,codon_table));

					k0 = offset3 - diff_position;
					low_flag = high_flag = FALSE;
					/* Search for stop codon */
					k1 = 0;
//					for(k=offset-1;k > 0;k--){
					for(k=offset3-1;k >= 0;k--){
						if(seq_aa1[k] == '*'){
							k1 = k+1;
							break;
						}
					}
					printf("Upstream stop codon at k1 = %ld in aa sequence\n",k1);
					fflush(stdout);

printf("offset=%ld,offset3=%ld,gaps=%ld\n",offset,offset3,gaps);
fflush(fout);
					/* Search for init codon */
					k0m = k0v = k0l = 0L;
					for(k = k0;k < offset;k = 2 * k0 - k){
						if(k < 0) low_flag = TRUE;
						else if(k >= offset3) high_flag = TRUE;
						if(low_flag && high_flag) break;
						if(k < k1 || k < 0 || k >= offset3){
							if(k >= k0) k += 1;
							continue;
						}
						printf("%ld ",k);
						diff2 = labs(k0 - k);
						if(diff2 < (long)std1 + gaps){
							if(seq_aa1[k]=='m'){
								if(k0m == 0L) k0m = k;
								if(k >= k0) k += 1;
								continue;
							}else if(seq_aa1[k]=='v'){
								if(k0v == 0L) k0v = k;
								if(k >= k0) k += 1;
								continue;
							}else if(seq_aa1[k]=='l'){
								if(k0l == 0L) k0l = k;
								if(k >= k0) k += 1;
								continue;
							}
						}
						if(k >= k0) k += 1;
					}

printf("k0=%ld,k1=%ld,k0m=%ld,k0v=%ld,k0l=%ld\n",k0,k1,k0m,k0v,k0l);
fflush(fout);
					if(k0m == 0L){
						if(k1 > 0L){
							if(k0v > 0) k0m = k0v;
							else if(k0l > 0) k0m = k0l;
							else {
								for(k=k1;k<offset3;k++){
									if(seq_aa1[k]=='m'|| seq_aa1[k]=='v' || seq_aa1[k]=='l'){
										k0m = k;
										break;
									}
								}
							}
						}
					}
					
					if((new_position = position - (offset3 - k0m)) < 1){
						if(labs(av1_position - new_position) > labs(position - av1_position)){
							k0m = -1L;
							if(print_flag) fprintf(fout,"A putative upstream init codon is located too far.\n");
						}
					}
					if(k0m > 0L || (k0m == 0L && k1 == 0L)){
						k=k0m;
						fwrite_fasta(fout,seqM[i],0,' ');
						if(seq_aa1[k] > 'Z' || k1 == 0L){
							new_position = position-(offset3-k);
							if(k1 == 0L){
								if(print_flag) fprintf(fout,"The ORF continues upstream of the available nucleic acid sequence for %s.\n",seqM[i]->ID);
							}
							if(print_flag) fprintf(fout,"New start=%ld average=%ld original position=%ld\n",new_position,av1_position,position);
							for(kk = k;kk < offset3;kk++){
								if(new_position - k + kk < 0L) continue;	/* check */
								(seqM[i]->seq)[new_position-k+kk] = seq_aa1[kk];
								flag = TRUE;	/* revised alignment will be printed */
							}
							if(print_flag){
								if(k0 < 0L){
									GapSpace(offset3-4);
									fprintf(fout,"<<<-%s\n",gap_space);
//									fprintf(fout,"<<<-%s\n",space_line(offset3-4,'-'));
								}else{
									GapSpace(k0);
									fprintf(fout,"%s",gap_space);
									for(kk=0;kk<gaps;kk++) fprintf(fout,"%c",'v');
									fprintf(fout,"\n");
//									fprintf(fout,"%s%s\n",space_line(k0,'-'),space_line(gaps,'v'));
								}
								fprintf(fout,"%s\n\n",seq_aa1);
								fflush(fout);
							}
							strcpy(seqM[i]->header,"revised");

							if(print_flag) fwrite_fasta(fout,seqM[i],1,' ');
							fprintf(fout,"\n");

							/* modifying nucleic acid sequence */
							if(mode == 2){
								new_size = seq_nuc[j]->seqlen + 3 * (offset3 - k);
								if((seq_nuc[j]->seq = (char*)realloc(seq_nuc[j]->seq,(3L + new_size) * sizeof(char)))==NULL){
									fprintf(stderr,"Memory allocation error in alnCorrect4.\n\n");
								}

								slide = 3 * (offset3 - k);
								seqlen2 = seq_nuc[j]->seqlen;
								for(np=seq_nuc[j]->seq + seqlen2;np >= seq_nuc[j]->seq;np--){
									*(np + slide) = *np;
								}
								for(np=seq_nuc[j]->seq + slide-1;np >= seq_nuc[j]->seq;np--){
									*np = '-';
								}

								seq_nuc[j]->seqlen = strlen(seq_nuc[j]->seq);
								if(strstr(seq_nuc[j]->header,"Revised:")!=NULL && strlen(seq_nuc[j]->header) < MAXLINE ){
									sprintf(memo,": %ld gaps added to the 5' terminus",slide);
									strcat(seq_nuc[j]->header,memo);
								}else{
									sprintf(memo,"Revised: %ld gaps added to the 5' terminus",slide);
									strcpy(seq_nuc[j]->header,memo);
								}
							}
							/* end of modifying nucleic sequence */
						}
					} else {
						if(print_flag){
							fprintf(fout,"In-frame upstream initiation codon was not found.\n");
							GapSpace(k0);
							fprintf(fout,"         %s",gap_space);
							for(kk=0;kk<gaps;kk++) fprintf(fout,"%c",'v');
							fprintf(fout,"\n");
//							fprintf(fout,"         %s%s\n",space_line(k0,'-'),space_line(gaps,'v'));
							fprintf(fout,"Frame 1: %s\n",seq_aa1);
							fprintf(fout,"Frame 2: %s\n",seq_aa2);
							fprintf(fout,"Frame 3: %s\n",seq_aa3);
							fprintf(fout,"\n");
						}
					}
					break;
				}
			}
			if(j == num_seq_nuc){
				fprintf(fout,"DNA sequence was not found for %s.\n",ID);
				no_DNA = TRUE;
			}
/********************/
			}	/* end of N-term extention */

//			num_seq = num_seq_bak;
		} else if(mat[num_seq * 9 + i] < num_seq_thr){
			if(print_flag) fprintf(fout,"Unaligned C-terminus\n");
		} else if(mat[num_seq * 12 + i] < num_seq_thr){
			if(print_flag) fprintf(fout,"Singular total length\n");
		}

		if(short_sequence || no_DNA){

			if(mode == 2 || (mode==1 && !strcmp(org_name,"all"))){
				if(print_flag){
					if(short_sequence) fprintf(fout,"Short sequence %s is being removed from the alignment\n",seqM[i]->ID);
					else if(no_DNA) fprintf(fout,"Aberrant sequence %s having no DNA sequence is being removed from the alignment\n",seqM[i]->ID);
					fprintf(fout,"\n");
					fflush(fout);
				}
				if(i == num_seq_new -1) num_seq_new -= 1L;
				else {
					for(ii=i; ii<num_seq_new-1;ii++){
						for(k=0;k<15;k++) mat[num_seq * k + ii] = mat[num_seq * k + ii + 1];
						strcpy(seqM[ii]->ID,seqM[ii+1]->ID);
						strcpy(seqM[ii]->header,seqM[ii+1]->header);
						strcpy(seqM[ii]->seq,seqM[ii+1]->seq);
					}
					flag = TRUE;
					num_seq_new -= 1L;
					i--;
				}
			}
			short_sequence = FALSE;
			no_DNA = FALSE;
		}

	}
	if(print_flag){
		fprintf(fout,"\n");
		fclose(fout);
	}else{
		if(mode==1 && strcmp(org_name,"all")){
			printf("There is no sequence of %s.\n",org_name);
			printf("If short sequences have been removed, a new alignment is printed.\n");
		}
	}

	if(mode == 2 || (mode==1 && !strcmp(org_name,"all"))){
		num_seq_bak = num_seq;
		num_seq = num_seq_new;
	}
	if(flag){
		write_clustal(newname,seqM);
		printf("\tA modified alignment is printed: %s.\n",newname);
	}
	if(mode == 2 || (mode==1 && !strcmp(org_name,"all"))) num_seq = num_seq_bak;

	return FALSE;
}
/******************************************************************/


/******************************************************************/
int alnStat(char *fname1, char *fname2, Seqfile **seqM, double thr)
/******************************************************************/
{
	FILE *fout;
	long i,j;
	long length;
	double total;
	double total2;
	double stdev;
	double av1,av2,av3,std1,std2,std3;
	double *mat;
	double num_seq_thr;
	int mat_size = 15 * num_seq + 1;
	long seqlen = seqM[0]->seqlen;
	long num_seq_local,seqlen_local;
	int mode=0;		/* 0 for stat, 1 for stat2 */

	if(thr == 1.0) mode = 0;
	else mode = 1;

	if((mat = (double*)calloc(mat_size, sizeof(double)))==NULL){
		fprintf(stderr,"Memory allocation error in alnStat.\n\n");
		return TRUE;
	}

	if((fout=fopen(fname2,"w"))==NULL){
		fprintf(stderr,"Cannot open file: %s\n\n",fname2);
		return TRUE;
	}

	fprintf(fout," Statistics of alignment: %s. Output of SISEQ\n\n",fname1);
	fprintf(fout," Number of sequences: %ld.\n",num_seq);

	total = 0.0;
	total2 = 0.0;
	for(i=0;i<num_seq;i++){
		for(j=0;j<seqlen;j++){
			if(seqM[i]->seq[j] != '-') break;
		}
		mat[num_seq * 0 + i] = (double)j;
		total +=  (double)j;
		total2 += (double)(j * j);
	}
	total /= (double)num_seq;
	stdev = (total2 / (double)num_seq) - total * total;
	stdev = sqrt(stdev);
	fprintf(fout," Start position: average = %6.3f stdev = %6.3f.\n",total,stdev);
	av1 = total;
	std1 = stdev;

	for(i=0;i<num_seq;i++){
		mat[num_seq * 1 + i] = (mat[num_seq * 0 + i] - av1 ) / std1;
		num_seq_local = 0L;
		seqlen_local = mat[num_seq * 0 + i];
		total = total2 = stdev = 0.0;
		for(j=0;j<num_seq;j++){
			if(fabs(mat[num_seq * 0 + j] - seqlen_local) <= std1){
				num_seq_local += 1L;
				total += (double)mat[num_seq * 0 + j];
				total2 += (double)(mat[num_seq * 0 + j] * mat[num_seq * 0 + j]);
			}
		}
		total /= (double)num_seq_local;
		stdev = (total2 / (double)num_seq_local) - total * total;
		stdev = sqrt(stdev);
		mat[num_seq * 6 + i] = (double)num_seq_local;
		mat[num_seq * 7 + i] = total;
		mat[num_seq * 8 + i] = stdev;
	}
	

	total = 0.0;
	total2 = 0.0;
	for(i=0;i<num_seq;i++){
		for(j=0;j<seqlen;j++){
			if(seqM[i]->seq[seqlen-j-1] != '-') break;
		}
		mat[num_seq * 2 + i] = (double)j;
		total +=  (double)j;
		total2 += (double)(j * j);
	}
	total /= (double)num_seq;
	stdev = (total2 / (double)num_seq) - total * total;
	stdev = sqrt(stdev);
	fprintf(fout," C-terminal vacant space: average = %6.3f stdev = %6.3f.\n",total,stdev);
	av2 = total;
	std2 = stdev;

	for(i=0;i<num_seq;i++){
		mat[num_seq * 3 + i] = (mat[num_seq * 2 + i] - av2) / std2;
		num_seq_local = 0L;
		seqlen_local = mat[num_seq * 2 + i];
		total = total2 = stdev = 0.0;
		for(j=0;j<num_seq;j++){
			if(fabs(mat[num_seq * 2 + j] - seqlen_local) <= std2){
				num_seq_local += 1L;
				total += (double)mat[num_seq * 2 + j];
				total2 += (double)(mat[num_seq * 2 + j] * mat[num_seq * 2 + j]);
			}
		}
		total /= (double)num_seq_local;
		stdev = (total2 / (double)num_seq_local) - total * total;
		stdev = sqrt(stdev);
		mat[num_seq * 9 + i] = (double)num_seq_local;
		mat[num_seq * 10 + i] = total;
		mat[num_seq * 11 + i] = stdev;
	}

	total = 0.0;
	total2 = 0.0;
	for(i=0;i<num_seq;i++){
		length = 0L;
		for(j=0;j<seqlen;j++){
			if(seqM[i]->seq[j] != '-') length += 1L;
		}
		mat[num_seq * 4 + i] = (double)length;
		total +=  (double)length;
		total2 += (double)(length * length);
	}
	total /= (double)num_seq;
	stdev = (total2 / (double)num_seq) - total * total;
	stdev = sqrt(stdev);
	fprintf(fout," Sequence length: average = %6.3f stdev = %6.3f.\n",total,stdev);
	av3 = total;
	std3 = stdev;

	for(i=0;i<num_seq;i++){
		mat[num_seq * 5 + i] = (mat[num_seq * 4 + i] - av3 ) / std3;
		num_seq_local = 0L;
		seqlen_local = mat[num_seq * 4 + i];
		total = total2 = stdev = 0.0;
		for(j=0;j<num_seq;j++){
			if(fabs(mat[num_seq * 4 + j] - seqlen_local) <= std3){
				num_seq_local += 1L;
				total += (double)mat[num_seq * 4 + j];
				total2 += (double)(mat[num_seq * 4 + j] * mat[num_seq * 4 + j]);
			}
		}
		total /= (double)num_seq_local;
		stdev = (total2 / (double)num_seq_local) - total * total;
		stdev = sqrt(stdev);
		mat[num_seq * 12 + i] = (double)num_seq_local;
		mat[num_seq * 13 + i] = total;
		mat[num_seq * 14 + i] = stdev;
	}

	fprintf(fout,"\n ID\tstart_av\tstart_dev\tend_av\tend_dev\tlength_av\tlength_dev");
	fprintf(fout,"\tStart_neighbor #seq\taverage\tstdev");
	fprintf(fout,"\tEnd_neighbor #seq\taverage\tstdev");
	fprintf(fout,"\tLength_neighbor #seq\taverage\tstdev");
	fprintf(fout,"\n");

	num_seq_thr = (double)num_seq * thr;
	for(i=0;i<num_seq;i++){
		if(mode==1){
			if(mat[num_seq * 6 + i] >= num_seq_thr && mat[num_seq * 9 + i] >= num_seq_thr && mat[num_seq * 12 + i] >= num_seq_thr){
				continue;
			}
		}
		fprintf(fout,"%s",seqM[i]->ID);
		for(j=0;j<15;j++){
			if(j == 6 || j == 9 || j == 12){
				fprintf(fout,"\t%ld",(long)mat[num_seq*j + i]);
			} else {
				fprintf(fout,"\t%6.3f",mat[num_seq*j + i]);
			}
		}
		fprintf(fout,"\n");
	}

	fprintf(fout,"\n");

	return FALSE;
}
/******************************************************/

/********************************/
int UnequalLength(Seqfile **seqM)
/********************************/
{
	long i;
	long seqlen;

	if(seqM==NULL){
		fprintf(stderr,"Null sequence.\n");
		return 1;
	}

	for(i=0;i<num_seq;i++){
		if(seqM[i]==NULL){
			fprintf(stderr,"Null sequence.\n");
			return 1;
		}
	}
	seqlen=seqM[0]->seqlen;

	for(i=0;i<num_seq;i++){
		if(seqlen!=seqM[i]->seqlen){
			return 1;
		}
	}

	return 0;
}



/******************************************************************/
void PrintScTable(FILE *fout,long N0,SimilarityCounts **sc1,int sw)
/******************************************************************/
{
	long i,j;
	double a=0.0;
	long b=0L;
	double av=0.0;
	double count=0.0;
	double x2=0.0;
	double std=0.0;

	for(i=1;i<N0;i++){
		for(j=0;j<i;j++){
			if(j==0) fprintf(fout,"%5ld: ",i+1);
//			if(sc1[i*N0+j]->total<=0.5) {
			if(sc1[i*N0+j]->total==0) {
				b=0L;
				fprintf(fout,"%5ld",b);
				continue;
			}
			if(sw==1){
				b=(long)sc1[i*N0+j]->match;
				fprintf(fout,"%5ld ",b);
			} 
			else if(sw==2) {
				b=(long)sc1[i*N0+j]->similar;
				fprintf(fout,"%5ld ",b);
			} 
			else if(sw==3) {
				a=(double)sc1[i*N0+j]->match/(double)sc1[i*N0+j]->total;
				fprintf(fout,"%5.3f ",a);
				av += a;
				x2 += a*a;
				count += 1.0;
			} 
			else if(sw==4){
				a=(double)sc1[i*N0+j]->similar/(double)sc1[i*N0+j]->total;
				fprintf(fout,"%5.3f ",a);
			}
			else fprintf(fout,"     0 ");
		}
		fprintf(fout,"\n");
	}
	fprintf(fout,"      ");
	for(i=0;i<N0-1;i++){
		fprintf(fout,"%5ld ",i+1);
	}
	fprintf(fout,"\n");

	av /= count;
	std = sqrt(x2/count - av * av);
	fprintf(fout,"av +- std = %5.3f +- %5.3f n = %ld\n",av,std,(long)count);

	return;
}

/*******************************************************************************************************/
int CountSimilarity(Seqfile *seq1,Seqfile *seq2,long N0,double st[30][30],SimilarityCounts **sc1,long pos)
/*******************************************************************************************************/
{
	long seqlen;
	long i,j,k;
	char a,b;
	double total=0.0;	/* counter */
	double match=0.0;	/* counter */
	double similar=0.0;	/* counter */

	if(seq1==NULL || seq2==NULL){
		fprintf(stderr,"Null sequence.\n");
		return 1;
	}
	else if(seq1->seq==NULL || seq2->seq==NULL){
		fprintf(stderr,"Null sequence.\n");
		return 1;
	}
	if((seqlen=seq1->seqlen)!=seq2->seqlen){

/* Now different lengths are allowed. Feb. 2007 */
		sc1[pos]->total=0L;
		sc1[pos]->match=0L;
		sc1[pos]->similar=0L;
#if 0
		sc1[pos]->total=0.0;
		sc1[pos]->match=0.0;
		sc1[pos]->similar=0.0;
#endif

/*		fprintf(stderr,"Length of the two sequences should be identical.\n");
		return 1;
*/
		return 0;
	}

	if(seqlen==0){
		fprintf(stderr,"Empty sequence.\n");
		for(i=1;i<N0;i++){
			for(j=0;j<i;j++){
				sc1[i*N0+j]->total=0L;
				sc1[i*N0+j]->match=0L;
				sc1[i*N0+j]->similar=0L;
#if 0
				sc1[i*N0+j]->total=0.0;
				sc1[i*N0+j]->match=0.0;
				sc1[i*N0+j]->similar=0.0;
#endif
			}
		}
		return 0;
	}
	
	total=0.0;
	match=0.0;
	similar=0.0;
	for(k=0;k<seqlen;k++) {
		a=seq1->seq[k];
		b=seq2->seq[k];
		if(islower(a)) a=toupper(a);
		if(islower(b)) b=toupper(b);
		if(!isalpha(a)) a=29;
		else a=a-'A';
		if(!isalpha(b)) b=29;
		else b=b-'A';
		similar+=st[(int)a][(int)b];
		if(a>=0 && a<27 && b>=0 && b<27) {
			total+=1.0;
			if(a==b) match+=1.0;
		}
	}

	sc1[pos]->total=(long)total;
	sc1[pos]->match=(long)match;
	sc1[pos]->similar=(long)similar;
#if 0
	sc1[pos]->total=total;
	sc1[pos]->match=match;
	sc1[pos]->similar=similar;
#endif

	return 0;
}

static int hl1=MINMEMSIZE; /* header_length */

/*****************************************************************************************************************/
int EraseCodingRegion(Seqfile *seq_in,Seqfile **seq_nuc,int c,Boolean retain_intron)
/*****************************************************************************************************************/
{
	int i;
	long k;
	long j0;
	int exons=0;
	Boolean circular_bak=FALSE; /* TRUE if circular. */
	static char *str1;
	static char *str2;
	static char *str3;
	char *header;
	char *ID;
	char *p;
	static char ACstr[MINMEMSIZE];
	Exon_record **exon;
	char comment[MAXLINE];

	circular_bak=circular;
	circular=seq_in->circular;


	if(str1==NULL || str2==NULL || str3==NULL){
		if((str1=(char*)calloc((hl1+2),sizeof(char)))==NULL){
			fprintf(stderr,"Out of memory in EraseCodingRegion1.\n");
			return 1;
		}
		if((str2=(char*)calloc((hl1+2),sizeof(char)))==NULL){
			fprintf(stderr,"Out of memory in EraseCodingRegion2.\n");
			return 1;
		}
		if((str3=(char*)calloc((hl1+2),sizeof(char)))==NULL){
			fprintf(stderr,"Out of memory in EraseCodingRegion3.\n");
			return 1;
		}
	}


	for(k=0;k<num_seq;k++){

		if(strcmp(seq_nuc[k]->ID,seq_in->ID)) continue;

		if(k==0 || (k>0 && hl1<(int)strlen(seq_nuc[k]->header))){
			hl1=strlen(seq_nuc[k]->header);
			if((str1=(char*)realloc(str1,(hl1+2)*sizeof(char)))==NULL){
				fprintf(stderr,"Out of memory in EraseCodingRegion4.\n");
				return 1;
			}
			if((str2=(char*)realloc(str2,(hl1+2)*sizeof(char)))==NULL){
				fprintf(stderr,"Out of memory in EraseCodingRegion5.\n");
				return 1;
			}
			if((str3=(char*)realloc(str3,(hl1+2)*sizeof(char)))==NULL){
				fprintf(stderr,"Out of memory in EraseCodingRegion6.\n");
				return 1;
			}
		}
		ID=seq_nuc[k]->ID;
		header=seq_nuc[k]->header;	
                strcpy(str1,seq_nuc[k]->header);
                strcpy(str2,str_shift(str1,' ')); /* str1 is gene name */
                strcpy(str3,str_shift(str2,' ')); /* str2 is exon number string */
                if(!num_check(str2)) exons=atoi(str2);
		else continue;

	/**************************************************/
		if((exon=(Exon_record**)calloc(exons+1,sizeof(Exon_record)))==NULL){
			fprintf(stderr,"Out of memory in EraseCodingRegion7\n");
			return 1;
		}
		for(i=0;i<exons;i++){
			if((exon[i]=(Exon_record*)calloc(1,sizeof(Exon_record)))==NULL){
				fprintf(stderr,"Out of memory in EraseCodingRegion8\n");
				return 1;
			}
		}
		if((p=strstr(str3,"AC="))!=NULL){
			subst(str3,';',' ');
			sscanf(p+3,"%s",ACstr);
		}
		strcpy(comment,"                                                  ");
		if(exon_list(seq_nuc[k]->header,exon,comment)){
			fprintf(stderr,"\nError in making exon list. Skipping %s %s.\n",ID,str1);
			continue;
		}
		
	/* erase sequence by filling int c, normally 0. */

		/* check sequence import. Normally, imported sequence is ORF, and therefore, may be erased. */
		if(strcmp(exon[0]->AC,"") && strcmp(getAC(exon[0]->AC),ACstr)){
			continue;
		}
		if(strcmp(exon[exons-1]->AC,"") && strcmp(getAC(exon[exons-1]->AC),ACstr)){
			continue;
		}
		
		for(i=0;i<exons;i++){
			for(j0=exon[i]->begin-1;j0<exon[i]->end;j0++){
				seq_in->seq[j0]=c;
			}
		}
		
		if(!retain_intron){
			if(!TransSplicing(exon,exons)){
				for(i=0;i<exons-1;i++){
					if(exon[0]->nc){	/* complementary */
						for(j0=exon[i]->begin-2;j0>exon[i+1]->end-1;j0--){
							seq_in->seq[j0]=c;
						}
					}
					else{			/* normal */
						for(j0=exon[i]->end;j0<exon[i+1]->begin-1;j0++){
							seq_in->seq[j0]=c;
						}
					}
				}
			}
		}

		free_exon(exon,exons);
		strcpy(str1,"");
		strcpy(str2,"");
		strcpy(str3,"");

		if(k%10==0) {
			putc(8,stdout);
			putc('.',stdout);
			fflush(stdout);
		}
	}
	circular=circular_bak;	
	return 0;
}


/************************************************************************/
int CountNonCoding(Seqfile *seq_in,long *NumSeq,int c)
/************************************************************************/
{
	long i=0L;
	char *p;
	long seqlen;
	Boolean flag=FALSE;

	if(seq_in==NULL) return 1;
	if(seq_in->seq==NULL) return 1;
	if(strcmp(seq_in->seq,"")==0) return 1;
	if((seqlen=seq_in->seqlen)<1) return 1;

	for(p=seq_in->seq;*p!='\0';p++){
		if(!flag && *p!=c){
			i+=1;
			flag=TRUE;
		}
		else if(flag && *p==c){
			flag=FALSE;
		}
	}
	
	*NumSeq=i;

	return 0;
}


/************************************************************************/
int SelectNonCoding(Seqfile *seq_in,Seqfile **seq_out,long *NumSeq,int c)
/************************************************************************/
{
	long i=0L;
	char *p,*p1;
	long seqlen;
	Boolean flag=FALSE;
	long sn=0L;
	Seqfile *s;
	enum data_type c1;

	if(seq_in==NULL || seq_out==NULL || NumSeq==NULL) return 1;
	if(seq_in->seq==NULL) return 1;
	if(strcmp(seq_in->seq,"")==0) return 1;
	if((seqlen=seq_in->seqlen)<1) return 1;

	for(p=p1=seq_in->seq;*p!='\0';p++){
		if(!flag && *p!=c){
			i+=1;
			p1=p;
			flag=TRUE;
		}
		else if(flag && *p==c){
			*p='\0';
			seq_in->seq=p+1;
			seq_out[i-1]->seq=p1;
			flag=FALSE;

/* debug */
#ifdef VERBOUS
			printf("%s\n",seq_out[i-1]->seq);
#endif
		}
	}
	if(*NumSeq==i){
		seq_in->seq=p;
		seq_out[i-1]->seq=p1;
	}
	else return 1;

	/* check data type */
	for(i=0,sn=0;i<*NumSeq;i++){
		seq_out[i]->seqlen=strlen(seq_out[i]->seq);
		c1=check_data_type(seq_out[i]);
#ifdef VERBOUS
	printf("data type is %s.\n",data_type_name[c1]);
#endif
		if(c1==DNA) {
			if(i!=sn){
				s=seq_out[sn];
				seq_out[sn]=seq_out[i];
			}
			sn+=1;
			continue;
		}
	}
	if(i!=sn) *NumSeq=sn;	

	return 0;
}


/***************************************************************/
int c_basecomp(Seqfile *seq_in,double basecomp[4])
/***************************************************************/
{
	if(seq_in==NULL) return TRUE;
	else if(seq_in->seqlen<=0) return TRUE;

	if(c_basecomp_seq(seq_in->seq,basecomp)) {
		fprintf(stderr,"Error in c_basecomp_seq.\n");
		return TRUE;
	}

	return FALSE;
}


/**************************************************/
int c_basecomp_seq(char *seq,double basecomp[4])
/**************************************************/
{
	long i;
	long bc[4];	/* base count */
	int a,j;
	double total=0.0;	/* total */
	long seqlen;

	if(seq==NULL) return TRUE;
	else seqlen=strlen(seq);

	for(j=0;j<4;j++) bc[j]=0L;
	for(i=0;i<seqlen;i++){
		a=seq[i];
		if(islower(a)) a=toupper(a);
		if(a=='T' || a=='U') bc[0]+=1L;
		else if(a=='C') bc[1]+=1L;
		else if(a=='A') bc[2]+=1L;
		else if(a=='G') bc[3]+=1L;
	}
	
	for(j=0;j<4;j++) basecomp[j]=(double)bc[j]/(double)seqlen;
	for(j=0;j<4;j++) total+=basecomp[j];

	if(total>1.01 || total <=0.95){
		fprintf(stderr,"Strange composition.\n");
		if(total > 0.5){
			for(j=0;j<4;j++) basecomp[j] /= total;
		}
		else return TRUE;
	}

	return FALSE;
}

/************************************************************/
int c_freq(Seqfile *seq_in,long *freq,long maxdist,int modeM)
/************************************************************/
{
	long i;
	int j=0;

	if(maxdist<=0) return -1;
	if(seq_in==NULL) return -1;
	else if(seq_in->seqlen<=0) return -1;

	if(!modeM) printf("Calculating dist=        ");
	for(i=1;i<=maxdist;i++){
		if(!modeM){
			for(j=0;j<8;j++) putc(8,stdout);
			printf("%8ld",i);
			fflush(stdout);
		}
		if(c_freq_i(seq_in,freq,i,modeM)){
			fprintf(stderr,"Error in calculating dinucleotide frequencies with i=: %ld.\n",i);
			return i;
		}
	}
	if(!modeM) printf("\n");

	return 0;	
}


/********************************************************/
int c_freq_i(Seqfile *seq_in,long *freq,long n,int modeM)
/********************************************************/
{
	long i,j;	/* sequence positions */
	long seqlen,maxlen;
	int k,l;	/* 0 for T, 1 for C, 2 for A, 3 for G. */
	Boolean circular_bak;
	long bpc[4][4];	/* base pair count */
	int a,b;	/* character for base */
	long total=0L;

	seqlen=seq_in->seqlen;
	if(n>=seqlen) {
		fprintf(stderr,"Distance %ld larger than sequence length %ld.\n",n,seqlen);
		return 1;
	}
	if(n<1) {
		fprintf(stderr,"Invalid distance %ld.\n",n);
		return 1;
	}
	circular_bak=circular;
	circular=seq_in->circular;

	for(k=0;k<4;k++){
		for(l=0;l<4;l++){
			bpc[k][l]=0;
		}
	}

	if(!circular) maxlen=seqlen-n;
	else maxlen=seqlen;
	for(i=0;i<maxlen;i++){
		a=seq_in->seq[i];
		if(islower(a)) a=toupper(a);
		if(a=='U') a='T';
		j=i+n;
		if(j>=seqlen) j=j-seqlen;
		b=seq_in->seq[j];
		if(islower(b)) b=toupper(b);
		if(b=='U') b='T';

		if(a=='T' && b=='T') bpc[0][0]+=1; 
		else if(a=='T' && b=='C') bpc[0][1]+=1; 
		else if(a=='T' && b=='A') bpc[0][2]+=1; 
		else if(a=='T' && b=='G') bpc[0][3]+=1; 
		else if(a=='C' && b=='T') bpc[1][0]+=1; 
		else if(a=='C' && b=='C') bpc[1][1]+=1; 
		else if(a=='C' && b=='A') bpc[1][2]+=1; 
		else if(a=='C' && b=='G') bpc[1][3]+=1; 
		else if(a=='A' && b=='T') bpc[2][0]+=1; 
		else if(a=='A' && b=='C') bpc[2][1]+=1; 
		else if(a=='A' && b=='A') bpc[2][2]+=1; 
		else if(a=='A' && b=='G') bpc[2][3]+=1; 
		else if(a=='G' && b=='T') bpc[3][0]+=1; 
		else if(a=='G' && b=='C') bpc[3][1]+=1; 
		else if(a=='G' && b=='A') bpc[3][2]+=1; 
		else if(a=='G' && b=='G') bpc[3][3]+=1; 
	}


	for(k=0;k<4;k++){
		for(l=0;l<4;l++){
			total+=bpc[k][l];
		}
	}
	
	if(n==1L && total!=maxlen){
		fprintf(stderr,"\nThe number of total site pairs %ld is different from\n",total);
		fprintf(stderr,"the number of maximum length %ld. But we continue.\n",maxlen);
		if(!modeM) printf("Calculating dist=        ");
	}  

	for(k=0;k<4;k++){
		for(l=0;l<4;l++){
			freq[16*(n-1)+4*k+l]+=bpc[k][l];
		}
	}
		
	circular=circular_bak;	
	return 0;	
}

/********************* end of file calcseq.c ***********************/

